import aquila_borg as borg
import numpy as np
from astropy.coordinates import SkyCoord
import astropy.units as apu
import configparser
import ast

import borg_velocity.utils as utils
from borg_velocity.utils import myprint
import borg_velocity.forwards as forwards
import borg_velocity.poisson_process as poisson_process
import borg_velocity.projection as projection

def radially_scatter(xtrue, frac_sig_x):
    """
    Radially perturb true positions by a fractional amount
    frac_sig_x, where the observer sits at x = (0, 0, 0).
    
    Args:
        :xtrue (np.ndarray): The true coordinates (Mpc/h), shape=(3,Nt)
        :frac_sig_x (float): The fractional uncertainty in the radial direction
        
    Returns:
        :xmeas (np.ndarray): The observed coordiantes (Mpc/h) of the tracers.
        :sigma_mu (float):  The uncertainty in the distance moduli of the tracers.
        
    """
    
    # Convert to RA, Dec, Distance
    rtrue = np.sqrt(np.sum(xtrue** 2, axis=0))  # Mpc/h
    c = SkyCoord(x=xtrue[0], y=xtrue[1], z=xtrue[2], representation_type='cartesian')
    RA = c.spherical.lon.degree
    Dec = c.spherical.lat.degree
    r_hat = np.array(SkyCoord(ra=RA*apu.deg, dec=Dec*apu.deg).cartesian.xyz)

    # Add noise to radial direction
    sigma_mu = 5. / np.log(10) * frac_sig_x
    mutrue = 5 * np.log10(rtrue * 1.e6 / 10)
    mumeas = mutrue + np.random.normal(size=len(mutrue)) * sigma_mu
    rmeas = 10 ** (mumeas / 5.) * 10 / 1.e6
    xmeas, ymeas, zmeas = rmeas[None,:] * r_hat

    xmeas = np.array([xmeas, ymeas, zmeas])
    
    return xmeas, sigma_mu


def borg_mock(s_hat, state, fwd_model, ini_file):
    
    config = configparser.ConfigParser()
    config.read(ini_file)
    
    nsamp = int(config['run']['nsamp'])
    
    # Run BORG density field
    output_density = np.zeros(fwd_model.getOutputBoxModel().N)
    fwd_model.forwardModel_v2(s_hat)
    fwd_model.getDensityFinal(output_density)
    state["BORG_final_density"][:] = output_density
    
    # Get growth rate
    cosmo = utils.get_cosmopar(ini_file)
    cosmology = borg.cosmo.Cosmology(cosmo)
    af = float(config['model']['af'])
    f = cosmology.gplus(af)  # dD / da
    f *= af / cosmology.d_plus(af)  # f = dlnD / dlna
    
    # Get velocity
    smooth_R = float(config['model']['smooth_R'])
    output_vel = forwards.dens2vel_linear(output_density, f, 
                          fwd_model.getOutputBoxModel().L[0], smooth_R)
    
    # Add bulk flow
    bulk_flow = np.array(ast.literal_eval(config['mock']['bulk_flow']))
    output_vel = output_vel + bulk_flow.reshape((3, 1, 1, 1))
    
    # Sample positions according to bias model
    bias_epsilon = float(config['model']['bias_epsilon'])
    R_max = float(config['mock']['R_max'])
    coord_true = [None] * nsamp
    coord_meas = [None] * nsamp
    sig_mu = [None] * nsamp
    for i in range(nsamp):
        frac_sig_x = float(config[f'sample_{i}']['frac_sig_rhMpc'])
        alpha = float(config[f'sample_{i}']['alpha'])
        Nt = int(config[f'sample_{i}']['Nt'])
        phi = (1. + output_density + bias_epsilon) ** alpha
        coord_true[i] = np.zeros((3,Nt))
        coord_meas[i] = np.zeros((3,Nt))
        nmade = 0
        while (nmade < Nt):
            ctrue = poisson_process.sample_3d(phi, Nt, 
                  fwd_model.getOutputBoxModel().L[0], fwd_model.getOutputBoxModel().xmin)
            cmeas, sig_mu[i] = radially_scatter(ctrue, frac_sig_x)
            
            # Only use tracers which lie within R_max
            r = np.sqrt(np.sum(ctrue**2, axis=0))
            m = r < R_max
            
            nnew = m.sum()
            if nmade + nnew > Nt:
                coord_true[i][:,nmade:] = ctrue[:,m][:,:Nt-nmade]
                coord_meas[i][:,nmade:] = cmeas[:,m][:,:Nt-nmade]
            else:
                coord_true[i][:,nmade:nmade+nnew] = ctrue[:,m]
                coord_meas[i][:,nmade:nmade+nnew] = cmeas[:,m]
            nmade = min(nmade + m.sum(), Nt)
        
    # Interpolate velocities to tracers
    interp_order = int(config['model']['interp_order'])
    vr_true = [None] * nsamp
    for i in range(nsamp):
        tracer_vel = projection.interp_field(
                                output_vel,
                                np.expand_dims(coord_true[i], axis=2),
                                fwd_model.getOutputBoxModel().L[0],
                                np.array(fwd_model.getOutputBoxModel().xmin),
                                interp_order
                                )
        vr_true[i] = projection.project_radial(
                                tracer_vel,
                                np.expand_dims(coord_true[i], axis=2),
                                np.zeros(3,)
                                )
        
    # Compute observed redshifts (including noise)
    sig_v = float(config['model']['sig_v'])
    cz_obs = [None] * nsamp
    for i in range(nsamp):
        rtrue = np.sqrt(np.sum(coord_true[i] ** 2, axis=1))
        zco = utils.z_cos(rtrue, cosmo.omega_m)
        cz_obs = utils.speed_of_light * zco + (1 + zco) * vr_true[i]
        cz_obs += np.random.normal(size=cz_obs.shape) * sig_v  # CHECK THIS LINE!!!!!
        
    # Add observational systematic due to incorrect distance estimate 
    # \mu -> \mu + 5 log10(A), or equivalently d -> A d
    for i in range(nsamp):
        muA = float(config[f'sample_{i}']['muA'])
        coord_meas[i] = coord_meas[i] * muA
        
    return coord_true, coord_meas, sig_mu, vr_true, cz_obs